from django.test import TestCase

from .calculator import Calculator

class TestCalculator(TestCase):
    def setUp(self):
        self.words = [
            'A','BE','DOG','HOG','HASH','BLACK','BEATLE','HATRED','MISSES','PANCAKE','RANDOMLY']
        self.c = Calculator(self.words)

    def test_get_relevant_words_1(self):
        position = 0
        segment_size = 1
        words = self.c.get_relevant_words(self.words, position, segment_size)
        self.assertEqual(len(words), 11)
        self.assertEqual(words, self.words)

    def test_get_relevant_words_2(self):
        position = 3
        segment_size = 2
        words = self.c.get_relevant_words(self.words, position, segment_size)
        self.assertEqual(len(words), 6)
        self.assertEqual(words, self.words[5:])

    def test_get_relevant_words_3(self):
        position = 1
        segment_size = 3 
        words = self.c.get_relevant_words(self.words, position, segment_size)
        self.assertEqual(len(words), 7)
        self.assertEqual(words, self.words[4:])

    def test_get_relevant_words_4(self):
        position = 1
        segment_size = 3
        words = self.c.get_relevant_words(self.words, position, segment_size, inverse=True)
        self.assertEqual(len(words), 7)
        self.assertEqual(words, [x[::-1] for x in self.words[4:]])

    def test_get_number_of_matches_1(self):
        segment = 'O'
        position = 1
        segment_size = 1
        relevant_words = self.c.get_relevant_words(self.words, position, segment_size)
        matches = self.c.get_number_of_matches(segment, position, segment_size, relevant_words)
        self.assertEqual(matches, 2)

    def test_get_number_of_matches_2(self):
        segment = 'AN'
        position = 1
        segment_size = 2
        relevant_words = self.c.get_relevant_words(self.words, position, segment_size)
        matches = self.c.get_number_of_matches(segment, position, segment_size, relevant_words)
        self.assertEqual(matches, 2)

    def test_split_into_segments_1(self):
        word = 'PASS'
        segment_size = 1
        segments = self.c.split_into_segments(word, segment_size)
        self.assertEqual([x for x in segments], ['P','A','S','S'])

    def test_split_into_segments_2(self):
        word = 'VICTORY'
        segment_size = 2
        segments = self.c.split_into_segments(word, segment_size)
        self.assertEqual([x for x in segments], ['VI','IC','CT','TO','OR','RY'])

    def test_split_into_segments_3(self):
        word = 'BLACK'
        segment_size = 3
        segments = self.c.split_into_segments(word, segment_size)
        self.assertEqual([x for x in segments], ['BLA','LAC','ACK'])

    def test_get_probability_by_segments(self):
        word = 'BANANA'
        segment_size = 2
        prob = self.c.get_probability_by_segments(self.words, word, segment_size)
        self.assertEqual(prob, 
            [
                ['BA',0],
                ['AN',2.0/9.0],
                ['NA',0],
                ['AN',0],
                ['NA',0]
            ]
        )

    def test_get_word_probability__regular_monophoneme(self):
        word = 'BANANA'
        segment_size = 1
        prob = self.c.get_word_probability(word, segment_size, 1)
        self.assertEqual([x[1] for x in prob], [3.0/11,4.0/10,2.0/9,0.0,0.0,0.0])

    def test_get_word_probability__regular_biphoneme(self):
        word = 'BANANA'
        segment_size = 2
        prob = self.c.get_word_probability(word, segment_size, 1)
        self.assertEqual([x[1] for x in prob], [0.0,2.0/9,0.0,0.0,0.0])

    def test_get_word_probability__regular_triphoneme(self):
        word = 'BEAGLE'
        segment_size = 3
        prob = self.c.get_word_probability(word, segment_size, 1)
        self.assertEqual([x[1] for x in prob], [1.0/9,0.0,0.0,0.0])

    def test_get_word_probability__inverse_monophoneme(self):
        word = 'FROG'
        segment_size = 1
        prob = self.c.get_word_probability(word, segment_size, 2)
        self.assertEqual([x[1] for x in prob], [0.0,1.0/9,2.0/10,2.0/11])

    def test_get_word_probability__inverse_biphoneme(self):
        word = 'BULLDOG'
        segment_size = 2
        prob = self.c.get_word_probability(word, segment_size, 2)
        self.assertEqual([x[1] for x in prob], [0.0,0.0,0.0,0.0,1.0/9,2.0/10])

    def test_get_word_probability__inverse_triphoneme(self):
        word = 'BULLDOG'
        segment_size = 3
        prob = self.c.get_word_probability(word, segment_size, 2)
        self.assertEqual([x[1] for x in prob], [0.0,0.0,0.0,0.0,1.0/9])

    def test_get_word_probability__combined_monophoneme(self):
        word = 'BANANANA'
        segment_size = 1
        prob = self.c.get_word_probability(word, segment_size, 3)
        self.assertEqual([x[1] for x in prob], [
            (3.0/11+0.0)/2,
            (4.0/10+1.0/2)/2,
            (2.0/9+1.0/5)/2,
            (0.0+1.0/6)/2,
            0.0,
            (0.0+3.0/9)/2,
            0.0,
            (0.0+1.0/11)/2
        ])

    def test_get_word_probability__combined_biphoneme(self):
        word = 'BLOG'
        segment_size = 2
        prob = self.c.get_word_probability(word, segment_size, 3)
        self.assertEqual([x[1] for x in prob], [
            (1.0/10+0.0)/2,
            0.0,
            (0.0+2.0/10)/2
        ])

    def test_get_word_probability__combined_triphoneme(self):
        word = 'HASHRED'
        segment_size = 3
        prob = self.c.get_word_probability(word, segment_size, 3)
        self.assertEqual([x[1] for x in prob], [
            (1.0/9+0.0)/2,
            (1.0/7+0.0)/2,
            0.0,
            0.0,
            (0.0+1.0/9)/2
        ])



class Calculator(object):
    def __init__(self, corpus):
        self.corpus = corpus

    def get_relevant_words(self, all_words, position, segment_size, inverse=False):
        """ 
        Filter relevant words. I.e. checking if one letter word matches biphoneme
        segment doesn't make sense. 
        """
        if inverse:
            return [x[::-1] for x in all_words if len(x) >= position + segment_size]
        return [x for x in all_words if len(x) >= position + segment_size]

    def split_into_segments(self, word, segment_size):
        """
        If segment size is equal to one, return list of letters. Otherwise, return
        appropriate segments.
        """
        if segment_size == 1:
            return list(word)
        return (word[x:x+segment_size] for x in range(len(word)-segment_size+1))

    def get_number_of_matches(self, segment, position, segment_size, relevant_words):
        """
        Find out number of matching segments in the set of relevant words.
        """
        return len(filter(lambda x: x[position:position+segment_size] == segment, relevant_words))

    def calc_proportion(self, matches, total):
        """
        Simple float division placed into a function to prevent Zero division error.
        """
        if matches and total:
            return float(matches) / float(total)
        return 0.0

    def get_probability_by_segments(self, corpus, word, segment_size, inverse=False):
        """
        Calculate probabilities for each segment.
        """
        proportions = []
        for position, segment in enumerate(self.split_into_segments(word, segment_size)):
            relevant_words = self.get_relevant_words(corpus, position, segment_size, inverse=inverse)
            matches = self.get_number_of_matches(segment, position, segment_size, relevant_words)
            if inverse:
                segment = segment[::-1]
            proportions.append([segment,self.calc_proportion(matches,len(relevant_words))])
        if inverse:
            proportions.reverse()
        return proportions

    def get_word_probability(self, word, segment_size, prob_type):
        """
        Call the above method to calculate different probability types depending on
        parameters passed. If 'combined probability' is asked for, this method calls
        itself recursively to calculate 'regular probability' and 'inverse probability'
        for each segment, and then returns the average of the two results.
        """
        if prob_type == 1:
            return self.get_probability_by_segments(self.corpus, word, segment_size)
        if prob_type == 2:
            return self.get_probability_by_segments(
                self.corpus, word[::-1], segment_size, inverse=True)
        return [
            [i[0],(i[1]+j[1])/2] for i,j in zip(
                self.get_word_probability(word, segment_size, 1),
                self.get_word_probability(word, segment_size, 2)
            )
        ]



from django import forms
from django.template.loader import render_to_string
from django.utils.encoding import force_unicode, force_text
from django.utils.safestring import mark_safe
from django.utils.html import conditional_escape, format_html

class IntegerRangeWidget(forms.widgets.Widget):
    template_name = 'integer-range-field.html'

    def render(self, name, value, attrs=None):
        context = {
            'name':name,
            'value':value or self.attrs['default'] or self.attrs['min_value'],
            'minValue':self.attrs['min_value'],
            'maxValue':self.attrs['max_value'],
        }
        return mark_safe(render_to_string(self.template_name, context))

class CustomRadioInput(forms.widgets.RadioChoiceInput):
    def render(self, name=None, value=None, attrs=None, choices=()):
        name = name or self.name
        value = value or self.value
        attrs = attrs or self.attrs
        label_for = ' for="%s"' % (self.attrs['id'])
        choice_label = conditional_escape(force_unicode(self.choice_label))
        return mark_safe(u'<div class="text-center">%s<label %s>%s</label></div>' % (self.tag(), label_for, choice_label))

class RadioCustomRenderer(forms.widgets.RadioFieldRenderer):
    choice_input_class = CustomRadioInput
    outer_html = '<ul class="row" {id_attr}>{content}</ul>'

    def render(self):
        b = unicode(12/len(self.choices))
        class_name = 'col-xs-12 col-md-' + b
        self.inner_html = '<li class="' + class_name + '">{choice_value}{sub_widgets}</li>'
        
        id_ = self.attrs.get('id')
        output = []
        for i, choice in enumerate(self.choices):
            choice_value, choice_label = choice
            if isinstance(choice_label, (tuple, list)):
                attrs_plus = self.attrs.copy()
                if id_:
                    attrs_plus['id'] += '_{}'.format(i)
                sub_ul_renderer = self.__class__(
                    name=self.name,
                    value=self.value,
                    attrs=attrs_plus,
                    choices=choice_label,
                )
                sub_ul_renderer.choice_input_class = self.choice_input_class
                output.append(format_html(
                    self.inner_html, choice_value=choice_value,
                    sub_widgets=sub_ul_renderer.render(),
                ))
            else:
                w = self.choice_input_class(self.name, self.value, self.attrs.copy(), choice, i)
                output.append(format_html(self.inner_html, choice_value=force_text(w), sub_widgets=''))
        return format_html(
            self.outer_html,
            id_attr=format_html(' id="{}"', id_) if id_ else '',
            content=mark_safe('\n'.join(output)),
        )



from django import forms

from .widgets import RadioCustomRenderer, IntegerRangeWidget

SEGMENT = (
    (1,'Monophoneme'),
    (2,'Biphoneme'),
    (3,'Triphoneme'),
)

PROPORTION = (
    (1,'Regular'),
    (2,'Inverted'),
    (3,'Combined'),
)

SUM_OR_AV = (
    (1,'Summed'),
    (2,'Averaged'),
)

class MainForm(forms.Form):
    segment_size = forms.ChoiceField(
        choices=SEGMENT, 
        initial=SEGMENT[0][0],
        widget=forms.RadioSelect(renderer=RadioCustomRenderer)
    )
    probability_type = forms.ChoiceField(
        choices=PROPORTION, 
        initial=PROPORTION[0][0],
        widget=forms.RadioSelect(renderer=RadioCustomRenderer)
    )
    summed_or_averaged = forms.ChoiceField(
        choices=SUM_OR_AV, 
        initial=SUM_OR_AV[1][0],
        widget=forms.RadioSelect(renderer=RadioCustomRenderer)
    )
    number_of_words_in_corpus = forms.CharField(
        widget=IntegerRangeWidget(attrs={
            'min_value':10,
            'max_value':9874,
            'default':9874,
        }))


class TextForm(forms.Form):
    text = forms.CharField(label='',widget=forms.Textarea(attrs={
            'placeholder':"Enter text here. Any character that doesn't belong to the Serbian alphabet (including white space) will be interpreted as a separator.",
            'spellcheck':'false'
        }))

    def __init__(self, *args, **kwargs):
        super(TextForm, self).__init__(*args, **kwargs)
        self.fields['text'].widget.attrs['class'] = 'col-xs-12 text-area'


from django.utils.deprecation import MiddlewareMixin
from debug_toolbar.middleware import DebugToolbarMiddleware

class AtopdedTo110DebugMiddleware(MiddlewareMixin, DebugToolbarMiddleware):
    """ formtools temporary workaround for Django 1.10 """
    pass



from django.shortcuts import render
from django.views.generic.edit import FormView
from formtools.wizard.views import SessionWizardView

from collections import OrderedDict

from .calculator import Calculator
from .forms import MainForm, TextForm, SEGMENT, PROPORTION, SUM_OR_AV
from .models import Word
from .utils import convert_all

class CalculatorWizard(SessionWizardView):
    form_list = [MainForm, TextForm]
    template_name = 'wizard_form.html'

    def done(self, form_list, **kwargs):
        # GET PARAMETERS FROM FORM
        probability_type = int(form_list[0].cleaned_data.get('probability_type'))
        segment_size = int(form_list[0].cleaned_data.get('segment_size'))
        sum_or_average = int(form_list[0].cleaned_data.get('summed_or_averaged'))
        num_of_words = int(form_list[0].cleaned_data.get('number_of_words_in_corpus'))

        # GET WORDS FROM FORM AND CONVERT THEM TO CYRILLIC UPPERCASE
        text = form_list[1].cleaned_data.get('text')
        if text:
            text = set(convert_all(text))

        # INSTANTIATE CALCULATOR AND PASS CORPUS (OR PART OF IT)
        c = Calculator([x['word'] for x in Word.objects.values()][:num_of_words])

        probs = {}
        for word in text:
            # FOR EVERY WORD GET SEGMENT AND SUMMED PROBABILITIES AND CALCULATE SUM
            by_segment = c.get_word_probability(word, segment_size, probability_type)
            summed_prob = float(sum(x[1] for x in by_segment))
            if sum_or_average == 1:
                # IF SUMMED PROBABILITY - PLACE SUMMED PROBABILITY INTO A DICT
                probs[word] = [summed_prob, by_segment]
            elif summed_prob > 0:
                # ELSE, CHECK IF SUM GREATER THAN ZERO AND AVERAGE IT, THEN PLACE INTO A DICT
                probs[word] = [summed_prob/(len(word)-segment_size+1), by_segment]
            else:
                # IF ZERO, PLACE ZERO INSTEAD
                probs[word] = [0.0, by_segment]

        return render(self.request, 'done.html', {
            # PUT PROBABILITIES INTO AN ORDERED DICT AND ORDER THEM DESCENDING
            'probability': OrderedDict(sorted(probs.items(), key=lambda x: x[1])),
            # PASS INPUT PARAMETERS TO CONTEXT
            'info': [[  'Segment size', OrderedDict(SEGMENT)[segment_size]],
                     [  'Probability type', OrderedDict(PROPORTION)[probability_type]],
                     [  'Summed or averaged', OrderedDict(SUM_OR_AV)[sum_or_average]],
                     [  'Number of words in corpus', num_of_words]
                    ],
        })


import re
from .mappings import lat2cir, bi2mono

def clean(text):
    """ Clean text and convert everything to uppercase. """
    text = text.upper()
    return re.sub(r'[~`!@#$%^&*()-=_+\{\}\[\]\\:\;\"\'\|\?\>\<\,\.\/QWXY]+', " ", text)

def convert2cir(text):
    """ Convert Latin script to Cyrilic."""
    for i, j in lat2cir.iteritems():
        text = text.replace(i, j)
    return text

def convert_to_monographeme(text):
    """ Convert bigraphemes to single Cyrilic."""
    for i, j in bi2mono.iteritems():
        text = text.replace(i, j)
    return text

def convert_all(text):
    """ Call all helper functions to convert Latin to Cyrilic uppercase and return list of strings. """
    return [x for x in convert_to_monographeme(convert2cir(clean(text))).split(' ') if x]
